architecture behavioral of arbiter is

    type read_ack_channel is null channel;
    type read_ack_ptr is access read_ack_channel;

    type read_request_type is
        record
            reader_id : integer;
            ack : read_ack_ptr;
        end record read_request_type;

    type write_ack_channel is null channel;
    type write_ack_ptr is access write_ack_channel;

    type write_request_type is
        record
            writeer_id : integer;
            ack : write_ack_ptr;
        end record write_request_type;

    type read_request_channel is
        channel of read_request_type;	-- includes reader id
    type read_finished_channel is null channel;

    type write_request_channel is
        channel of write_request_type;	-- includes writer id
    type write_finished_channel is null channel;
    
    channel read_request : read_request_channel;
    channel read_finished : read_finished_channel;
    channel write_request : write_request_channel;
    channel write_finished : write_finished_channel;

begin

    access_controller : process is
    
        variable number_of_readers, number_of_writers : natural := 0;
        variable read_request_info : read_request_type;
        variable write_request_info : write_request_type;
    
    begin
        select
            when number_of_writers = 0 =>
            	receive read_request_info from read_request;
            	number_of_readers := number_of_readers + 1;
            	-- acknowledge read request
                send to read_request_info.ack.all;
        or
       	    receive from read_finished;
       	    number_of_readers := number_of_readers - 1;
        or
            when number_of_readers = 0 and number_of_writers = 0 =>
        	receive write_request_info from write_request;
        	number_of_writers := number_of_writers + 1;
        	-- acknowledge write request
                send to write_request_info.ack.all;
        or
            receive from write_finished;
            number_of_writers := number_of_writers - 1;
        end select; 
    end process access_controller;

end architecture behavioral;
